
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/namei.h>
#include <sys/kernel.h>
#include <sys/stat.h>
#include <sys/buf.h>
#include <sys/mount.h>
#include <sys/vnode.h>

#if __FreeBSD__==3
#include <miscfs/specfs/specdev.h>
#endif

#include <miscfs/fifofs/fifo.h>
#include <sys/malloc.h>
#include <sys/dirent.h>
#include <sys/unistd.h>
#include <sys/dirent.h>

#include <vm/vm.h>
#include <vm/vm_zone.h>
#include <vm/vnode_pager.h>

#include "ods2.h"

static int ods2_setattr __P((struct vop_setattr_args *));
static int ods2_access __P((struct vop_access_args *));
static int ods2_getattr __P((struct vop_getattr_args *));
static int ods2_pathconf __P((struct vop_pathconf_args *));

static int ods2_read __P((struct vop_read_args *));
static int ods2_readdir __P((struct vop_readdir_args *));

static int ods2_strategy __P((struct vop_strategy_args *));
static int ods2_print __P((struct vop_print_args *));
static int ods2_getpages __P((struct vop_getpages_args *));
static int ods2_putpages __P((struct vop_putpages_args *));
static int ods2_inactive __P((struct vop_inactive_args *)); 
static int ods2_reclaim __P((struct vop_reclaim_args *));
static int ods2_ioctl __P((struct vop_ioctl_args *));

mode_t ods2_mode_to_unix __P((unsigned short int));
void ods2_time_to_unix __P((struct vms_time_t *vmstime,
                            struct timespec *unixtime));
void make_version_string __P(( unsigned short int version,
                          char *version_string,
                          unsigned int *version_length ));
int extract_dir_record __P(( struct mount *mp,
                             unsigned int lowercase,
                             unsigned int flags,
                             int *offset,
                             unsigned char *buffer,
                             struct uio *uio ));
int extract_dir_block __P((struct mount *mp,
                           unsigned lowercase,
                           unsigned flags,
                           unsigned char *p,
                           struct uio *uio,
                           int *offset));

int ods2_read(ap)
    struct vop_read_args /* {
        struct vnode *a_vp;
        struct uio *a_uio;
        int a_ioflag;
        struct ucred *a_cred;
    } */ *ap;
{
    struct vnode *vp = ap->a_vp;
    struct uio *uio = ap->a_uio;
    struct ucred *cred = ap->a_cred;
    struct ods2_node *ip = VTOODS2(vp);
    struct ods2_mount *imp = MOUNTTOODS2MOUNT(ip->mount);
    off_t offset = uio->uio_offset;
    int original_byte_count = uio->uio_resid;
    unsigned vbn;
    unsigned lbn;
    unsigned offset_into_block;
    unsigned nbytes_to_move;
    int nblocks;
    int error;
    int bytes_left_in_file;
    unsigned char block[ODS2_BLOCK_SIZE];

    /*
     * Check end of file.
     * Convert the byte offset into a VBN
     * Map the VBN to LBN
     * Call ods2_bread on the underlying device using the LBN
     * Move data into uio structure
     * That's all folks
     */

    /* Invalid read request, byte offset into file < 0 */

    if (offset < 0)
    {
        if (0)
            printf("ods2_read[offset < 0]\n");

        return (EINVAL);
    }

    /* User doesn't want any bytes */

    if (original_byte_count == 0)
    {
        if (0)
            printf("ods2_read[original byte count = 0, no bytes requested]\n");

        return (0);
    }

    /* Trying to read past eof */
 
    if (offset >= ip->i_size)
    {
        if (0)
            printf("ods2_read[reading past eof]\n");

        return 0;
    }

    bytes_left_in_file = ip->i_size - offset;

    /* Now, the real work */

    /*
     * vbn's are 1 based
     */

    vbn = offset / ODS2_BLOCK_SIZE + 1;

    offset_into_block = offset % ODS2_BLOCK_SIZE;

    nblocks = ( original_byte_count + ODS2_BLOCK_SIZE - 1 ) / ODS2_BLOCK_SIZE;

    while (original_byte_count && bytes_left_in_file)
    {
        /*
         * Convert this vbn to lbn and read it.
         */

        error = VOP_BMAP( vp, vbn, NULL, &lbn, NULL, NULL );
        if (error)
            return error;

        if (0)
            printf("ods2_read[offset=%d, vbn=%d, lbn=%d]\n",
                   (int) offset, vbn, lbn);

        error = ods2_bread(imp->ods2_devvp, &imp->geometry, lbn, cred, block);
        if (error)
            return error;

        nbytes_to_move = ODS2_BLOCK_SIZE - offset_into_block;
        if ( original_byte_count < nbytes_to_move )
            nbytes_to_move = original_byte_count;
        if (bytes_left_in_file < nbytes_to_move)
            nbytes_to_move = bytes_left_in_file;

        uiomove(block + offset_into_block, nbytes_to_move, uio);

        original_byte_count -= nbytes_to_move;
        bytes_left_in_file -= nbytes_to_move;
        vbn++;
        offset_into_block = 0;
    }

    return 0;
}

int extract_dir_record( struct mount *mp,
                        unsigned int lowercase,
                        unsigned int flags,
                        int *offset,
                        unsigned char *buffer,
                        struct uio *uio )
{
    char vms_name[86];      /* MAX VMS NAME LENGTH */ /* XXX symbolic please */
    char version_string[5]; /* MAX VMS VERSION LENGTH */
    int error;
    int original_offset;
    unsigned pad_count;
    unsigned namestring;
    unsigned number_bytes;
    unsigned overhead;
    unsigned char *p = buffer;
    unsigned char *versionp;
    unsigned int i;
    unsigned version_length;
    unsigned full_length;
    struct dir_record_t *dir_record;
    struct dir_entry_t *dir_entry;
    struct dirent dirent;
    struct file_header_t *fh2;

    original_offset = *offset;

    dir_record = (struct dir_record_t *) buffer;

    pad_count = dir_record->dir_b_namecount % 2;
    namestring = dir_record->dir_b_namecount + pad_count;

    /* From Kirby: */

    overhead = DIR_S_DIR_RECORD_OVERHEAD;
    number_bytes = ( dir_record->dir_w_size - overhead - namestring );
    (*offset) += overhead + namestring;
    p = dir_record->dir_t_name + namestring;

    bcopy(dir_record->dir_t_name, vms_name, dir_record->dir_b_namecount);

    if (lowercase)
    {
        char *p;
        int i;

        for (i=0,p=vms_name; i < dir_record->dir_b_namecount; i++,p++)
            if (*p >= 'A' && *p <= 'Z') (*p) += ' ';
    }

    versionp = vms_name + dir_record->dir_b_namecount;
    bcopy(".", versionp, 1); /* . instead of ';' */
    versionp++;
    
    for (i=0; i < number_bytes; )
    {
        unsigned extension = 1;                /* Assume extension requested */
        dir_entry = (struct dir_entry_t *) p;
    
        dirent.d_fileno = FILE_NUMBER(dir_entry->dir_w_fid);

        if ((flags & ODS2MNT_FAST_NODIREXT) && (dir_entry->dir_w_version == 1))
        {
            /*
             * use dir_record version because vms_name may have been lowercased
             */

            char *p = dir_record->dir_t_name + dir_record->dir_b_namecount - 4;
            extension = bcmp(p, ".DIR", 4);
        }
        else if ((flags & ODS2MNT_NODIREXT) && (dir_entry->dir_w_version == 1))
        {
            /*
             * VMS directories are ALWAYS .DIR;1.  If .DIR;2, then VMS
             * doesn't treat it as a directory.  So, if the file is NOT
             * .DIR;1, then don't bother getting the file header.
             *
             * Go fetch this file header.  We do this to ensure its a
             * directory.  Note that we don't _vget it, we just read its
             * header.
             */

            char *p = dir_record->dir_t_name + dir_record->dir_b_namecount - 4;
            unsigned not_dirext = bcmp(p, ".DIR", 4);

            if (!not_dirext)
            {
                error = ods2_read_file_header( mp, dirent.d_fileno, &fh2 );
                if ( error )
                    return error;

                extension = !(fh2->fh2_l_filechar & FCH_M_DIRECTORY);

                FREE( fh2, M_ODS2FSFH );
            }
        }

        if (extension)
        {
            make_version_string(dir_entry->dir_w_version,
                                version_string,
                                &version_length);
    
            bcopy(version_string, versionp, version_length);
 
            full_length = dir_record->dir_b_namecount + 1 + version_length;
        }
        else
        {
            /*
             * VMS directories are ALWAYS .DIR, which is 4 characters.
             */

            full_length = dir_record->dir_b_namecount - 1 - 3;
        } 
    
        dirent.d_type = DT_REG; /* XXX: Go read the header :( */
        dirent.d_namlen = full_length;
        bcopy(vms_name, dirent.d_name, full_length);
        dirent.d_name[full_length] = 0; /* Must be null terminated */
        dirent.d_reclen = GENERIC_DIRSIZ(&dirent);

        if (uio->uio_resid < dirent.d_reclen)
            return -1;

        if (*offset >= original_offset)
        {
            if (0)
                printf("moving %s into uio at uio.offset=%d\n",
                       dirent.d_name,
                       (int) uio->uio_offset);

            error = uiomove((caddr_t) &dirent, dirent.d_reclen, uio);
            if (error) return error;

            (*offset) += DIR_S_ENTRY_FID;
        }
 
        p += DIR_S_ENTRY_FID;
        i += DIR_S_ENTRY_FID;
    }
    
    return 0;
}   

int extract_dir_block(struct mount *mp,
                      unsigned lowercase,
                      unsigned flags,
                      unsigned char *p,
                      struct uio *uio,
                      int *offset)
{
    struct dir_record_t *dir_record;
    int status;
    unsigned end_of_block;

    dir_record = (struct dir_record_t *) p;
    end_of_block = dir_record->dir_w_size == -1;

    if (0)
        printf("before extract_dir_block, offset=%d\n", *offset);

    while (!end_of_block)
    {
        status = extract_dir_record( mp,
                                     lowercase,
                                     flags,
                                     offset,
                                     p,
                                     uio );

        if ( status ) return status;

        /*
         * The record length of course doesn't include itself.
         */

        p += dir_record->dir_w_size + sizeof(dir_record->dir_w_size);
        dir_record = (struct dir_record_t *) p;
        end_of_block = dir_record->dir_w_size == -1;
    }

    (*offset) += ODS2_BLOCK_SIZE - (*offset) % ODS2_BLOCK_SIZE;

    if (0)
        printf("after extract_dir_block, offset=%d\n", *offset);

    return 0;
}

int ods2_readdir(ap)
    struct vop_readdir_args /* {
            struct vnode *a_vp;
            struct uio *a_uio;
            struct ucred *a_cred;
            int *a_eofflag;
            int *a_ncookies;
            u_long *a_cookies;
    } */ *ap;
{
    struct ucred *cred;
    struct vnode *vp;
    struct ods2_node *ip;
    struct uio *uio;
    struct uio block_uio;
    struct iovec block_iov;
    unsigned offset;
    int nbytes;
    int nbytes_read;
    int original_nbytes;
    unsigned total_bytes;
    int error;
    int block_aligned_offset;
    unsigned char *directory_block_buffer;
    struct dirent dirent;
    char *dot = ".";
    char *dotdot = "..";
    char *pound = "#";
    struct ods2_mount *mp;
    unsigned lowercase;
    unsigned flags;

    cred = ap->a_cred;
    vp = ap->a_vp;
    ip = VTOODS2(vp);
    mp = MOUNTTOODS2MOUNT(ip->mount);
    lowercase = mp->flags & ODS2MNT_LOWERCASE;
    flags = mp->flags;
    uio = ap->a_uio;
    offset = uio->uio_offset;
    original_nbytes = nbytes = uio->uio_resid;

    /*
     * Check eof
     */

    if (offset > ip->i_size)
    {
        *ap->a_eofflag = 1;
        return 0;
    }

    if (offset + nbytes >= ip->i_size)
    {
        original_nbytes = nbytes = ip->i_size - offset;
        *ap->a_eofflag = 1;
    }

    /*
     * user buffer must hold at least on entry
     */

    if (nbytes < sizeof(struct dirent))
        return EINVAL;

    if (0)
        printf("ods2_readdir[inum=%d, offset=%d, nbytes=%d]\n",
               ip->i_num, offset, nbytes);

    /*
     * VMS does not have "." and ".." in its directory.  As a result, these
     * entries have to be put in "by hand".  We do this here.  ALL offsets
     * into the directory biased by the size of "." + "..".
     *
     * . points to us - that's easy because we have the file header in vp
     * .. points to our parent.  That might be hard to do if the back-link
     *    pointer says we have no parent.  But, if we have no parent, then
     *    how is the user performing a directory?  No one could have gotten
     *    to us anyway!
     */

    /*
     * Calculate the VBN of offset
     */

    total_bytes = 0;
    nbytes_read = 0;

    /*
     * If offset < 512, then add in . and .. and adjust offset to 0
     * If offset > 512, then don't add in . and .. and adjust offset - 512
     */

    if (offset < ODS2_BLOCK_SIZE)
    {
        if (offset < sizeof(struct dirent))
        {
            if (0)
                printf("ods2_readdir[\tadding ., uio.resid=%d]\n",
                       uio->uio_resid);

            /* first "." */

            dirent.d_fileno = ip->i_num;
            dirent.d_type = DT_DIR;
            dirent.d_namlen = 1;
            bcopy(dot, dirent.d_name, 1);
            dirent.d_name[1] = 0; /* Must be null terminated */
            dirent.d_reclen = GENERIC_DIRSIZ(&dirent);

            error = uiomove((caddr_t) &dirent, dirent.d_reclen, uio);
            if (error) return error;
  
            total_bytes += dirent.d_reclen;
            nbytes -= dirent.d_reclen;
        }

        if (nbytes > 0)
        {
            if (0)
                printf("ods2_readdir[\tadding .., uio.resid=%d]\n",
                       uio->uio_resid);

            /* now ".." */

/* XXX: if backlink is 0 => no directory, then map this file to the
        SYSLOST directory - note, must "create" SYSLOST.DIR (or use if
        exists) during mount. */

            dirent.d_fileno = FILE_NUMBER(ip->file_header->fh2_w_backlink);
            dirent.d_type = DT_DIR;
            dirent.d_namlen = 2;
            bcopy(dotdot, dirent.d_name, 2);
            dirent.d_name[2] = 0; /* Must be null terminated */
            dirent.d_reclen = GENERIC_DIRSIZ(&dirent);

            if (nbytes < dirent.d_reclen)
                return 0;

            error = uiomove((caddr_t) &dirent, dirent.d_reclen, uio);
            if (error) return error;
  
            total_bytes += dirent.d_reclen;
            nbytes -= dirent.d_reclen;
        }

        if ( ip->i_num == MFD_FILE_NUMBER && nbytes > 0 )
        {
            dirent.d_fileno = mp->fid_directory_fid;
            dirent.d_type = DT_DIR;
            dirent.d_namlen = 1;
            bcopy(pound, dirent.d_name, 1);
            dirent.d_name[1] = 0;
            dirent.d_reclen = GENERIC_DIRSIZ(&dirent);

            if (nbytes < dirent.d_reclen)
                return 0;

            error = uiomove((caddr_t) &dirent, dirent.d_reclen, uio);
            if (error) return error;

            total_bytes += dirent.d_reclen;
            nbytes -= dirent.d_reclen;
        }

        offset = 0;
        nbytes_read = ODS2_BLOCK_SIZE;
        nbytes = original_nbytes - ODS2_BLOCK_SIZE;

        /*
         * AHHHHHH.  If nbytes == 0 - because we have an empty directory - that
         *           is a directory with NOTHING, not even -1 record length -
         *           then we'll never enter the "for" loop.  The only time that
         *           can happen is for the fid directory.  So, we need to adjust
         *           the offset...
         */

        if ( nbytes == 0 )
        {
            *ap->a_eofflag = 1;    /* make sure */
            uio->uio_offset = ODS2_BLOCK_SIZE;
            return 0;
        }

        if (0)
            printf("ods2_readdir[\tafter adding . and .., uio.resid=%d, offset=%d]\n",
                   uio->uio_resid, (int) uio->uio_offset);
    }

    block_uio.uio_iov = &block_iov;
    block_uio.uio_rw = UIO_READ;
    block_uio.uio_segflg = UIO_SYSSPACE;
    block_uio.uio_iovcnt = 1;
    block_uio.uio_resid = ODS2_BLOCK_SIZE;
    block_uio.uio_procp = uio->uio_procp;

    MALLOC(directory_block_buffer, caddr_t, ODS2_BLOCK_SIZE, M_TEMP, M_WAITOK);

    while (nbytes > 0 && offset < ip->i_size - ODS2_BLOCK_SIZE)
    {
        block_aligned_offset = (offset / ODS2_BLOCK_SIZE) * ODS2_BLOCK_SIZE;

        if (0)
            printf("ods2_readdir[actual reads: ba offset=%d, nbytes=%d]\n",
                   block_aligned_offset, nbytes);

        block_uio.uio_resid = ODS2_BLOCK_SIZE;
        block_uio.uio_offset = block_aligned_offset;

        block_iov.iov_len = ODS2_BLOCK_SIZE;
        block_iov.iov_base = directory_block_buffer;

        error = VOP_READ(vp, &block_uio, 0, cred);
        if (error)
        {
            FREE(directory_block_buffer, M_TEMP);
            return error;
        }

if (0)
{
    short int *rl = (short int *) directory_block_buffer;
    printf( "directory_block_buffer=%d\n", *rl);
}

        /*
         * Extract out the data from this directory block.  We do this by
         * extracting a directory record, then the directory entries.
         */

        error = extract_dir_block(ip->mount,
                                  lowercase,
                                  flags,
                                  directory_block_buffer,
                                  uio,
                                  &offset);
        if (error < 0) break;
        if (error)
        {
            FREE(directory_block_buffer, M_TEMP);
            return error;
        }

        nbytes = uio->uio_resid;
    }

/*
 * XXX: Align uio->uio_offset to directory entry boundary
 *      We will have to parse through the block until we find the entry
 *      that matches the offset - keep track of last directory record.
 */

    FREE(directory_block_buffer, M_TEMP);

    uio->uio_offset = offset;

    return 0;
}

/*
 * Setattr call. Only allowed for block and character special devices.
 */
int
ods2_setattr(ap)
	struct vop_setattr_args /* {
		struct vnodeop_desc *a_desc;
		struct vnode *a_vp;
		struct vattr *a_vap;
		struct ucred *a_cred;
		struct proc *a_p;
	} */ *ap;
{
    return (EROFS);
}

/*
 * Check mode permission on inode pointer. Mode is READ, WRITE or EXEC.
 * The mode is shifted to select the owner/group/other fields. The
 * super user is granted all permissions.
 */
/* ARGSUSED */
static int
ods2_access(ap)
	struct vop_access_args /* {
		struct vnode *a_vp;
		int  a_mode;
		struct ucred *a_cred;
		struct proc *a_p;
	} */ *ap;
{
	struct vnode *vp = ap->a_vp;
	struct ods2_node *ip = VTOODS2(vp);
	struct ucred *cred = ap->a_cred;
	mode_t mask, mode = ap->a_mode;
	gid_t *gp;
	int i;
        int fileprotection;

if (0)
printf("ods2_access[starting]\n");

	/* No writes */
	if (mode & VWRITE)
		return (EROFS);

	/* User id 0 always gets access. */
	if (cred->cr_uid == 0)
		return (0);

	mask = 0;

       fileprotection = OWNER_PROTECTION(ip->file_header->fh2_w_fileprot);

/* XXX: The full UIC must match - VMS semantics. */

	/* Otherwise, check the owner. */
	if (cred->cr_uid == ip->file_header->fh2_l_fileowner.uid) {
		if (mode & VEXEC)
			mask |= ODS2_X;
		if (mode & VREAD)
			mask |= ODS2_R;
		if (mode & VWRITE)
			mask |= ODS2_W | ODS2_D;
		return ((fileprotection & mask) != mask ? 0 : EACCES);
	}

	/* Otherwise, check the groups. */

        fileprotection = GROUP_PROTECTION(ip->file_header->fh2_w_fileprot);

	for (i = 0, gp = cred->cr_groups; i < cred->cr_ngroups; i++, gp++)
		if (ip->file_header->fh2_l_fileowner.gid == *gp) {
			if (mode & VEXEC)
				mask |= ODS2_X;
			if (mode & VREAD)
				mask |= ODS2_R;
			if (mode & VWRITE)
				mask |= ODS2_W | ODS2_D;
			return ((fileprotection & mask) != mask ?  0 : EACCES);
		}

	/* Otherwise, check everyone else. */

        fileprotection = WORLD_PROTECTION(ip->file_header->fh2_w_fileprot);

	if (mode & VEXEC)
		mask |= ODS2_X;
	if (mode & VREAD)
		mask |= ODS2_R;
	if (mode & VWRITE)
		mask |= ODS2_W | ODS2_D;
	return ((fileprotection & mask) != mask ? 0 : EACCES);
}

mode_t 
ods2_mode_to_unix(unsigned short int fileprot)
{
    mode_t unix = 0;

    int owner_protection = OWNER_PROTECTION(fileprot);
    int group_protection = GROUP_PROTECTION(fileprot);
    int world_protection = WORLD_PROTECTION(fileprot);

    if ((owner_protection & ODS2_R) == 0) unix |= S_IRUSR;
    if ((owner_protection & ODS2_X) == 0) unix |= S_IXUSR;
    if ((owner_protection & ODS2_W) == 0 ||
        (owner_protection & ODS2_D) == 0)
        unix |= S_IWUSR;

    if ((group_protection & ODS2_R) == 0) unix |= S_IRGRP;
    if ((group_protection & ODS2_X) == 0) unix |= S_IXGRP;
    if ((group_protection & ODS2_W) == 0 ||
        (group_protection & ODS2_D) == 0)
        unix |= S_IWGRP;

    if ((world_protection & ODS2_R) == 0) unix |= S_IROTH;
    if ((world_protection & ODS2_X) == 0) unix |= S_IXOTH;
    if ((world_protection & ODS2_W) == 0 ||
        (world_protection & ODS2_D) == 0)
        unix |= S_IWOTH;

    return unix;
}

void
ods2_time_to_unix(struct vms_time_t *vmstime, struct timespec *unixtime)
{
    /* VMS time is tenths of microseconds from 17-NOV-1858 00:00:00.00 */

    /* subtract the time from base vms to 1970 */
    /* convert result into seconds from 1970 by dividing by 100,000 */
    /* there are 100,000 tenths of microseconds in one second */

    unixtime->tv_sec = 901234567;
    unixtime->tv_nsec = 0;
}

static int
ods2_getattr(ap)
    struct vop_getattr_args /* {
        struct vnode *a_vp;
        struct vattr *a_vap;
        struct ucred *a_cred;
        struct proc *a_p;
    } */ *ap;

{
    struct vnode *vp = ap->a_vp;
    register struct vattr *vap = ap->a_vap;
    register struct ods2_node *ip = VTOODS2(vp);
    struct fi2_t *fi2;

if(0)
printf("ods2_getattr[starting]\n");

#if __FreeBSD__==3

    vap->va_fsid      = ip->i_dev;

#else

    vap->va_fsid      = dev2udev(ip->i_dev);

#endif

    vap->va_fileid    = ip->i_num;

    vap->va_mode      = ods2_mode_to_unix(ip->file_header->fh2_w_fileprot);
    vap->va_nlink     = 1;
    vap->va_uid	      = ip->file_header->fh2_l_fileowner.uid;
    vap->va_gid	      = ip->file_header->fh2_l_fileowner.gid;

    fi2               = FI2_AREA(ip->file_header);
    ods2_time_to_unix(&fi2->fi2_q_revdate, &vap->va_atime);
    vap->va_mtime     = vap->va_atime;
    vap->va_ctime     = vap->va_atime;

/*    vap->va_rdev      = vp->va_rdev; */ /* XXX: is this correct? */

    vap->va_flags     = 0;
    vap->va_gen       = 1;
    vap->va_blocksize = ODS2_BLOCK_SIZE; /* ALWAYS (512) bytes */
    vap->va_bytes     = (u_quad_t) ip->i_size;
    vap->va_size      = (u_quad_t) ip->i_size;

    vap->va_type      = vp->v_type;
    vap->va_filerev   = 0;

if(0)
printf("ods2_getattr[\tending]\n");

    return (0);
}

/*
 * Calculate the logical to physical mapping if not done already,
 * then call the device strategy routine.
 */
/*
 * XXX: Is this even needed?
 */
static int
ods2_strategy(ap)
	struct vop_strategy_args /* {
		struct buf *a_vp;
		struct buf *a_bp;
	} */ *ap;
{
	register struct buf *bp = ap->a_bp;
	register struct vnode *vp = bp->b_vp;
	register struct ods2_node *ip;
	int error;

if (0)
printf("ods2_strategy[starting]\n");

	ip = VTOODS2(vp);
	if (vp->v_type == VBLK || vp->v_type == VCHR)
		panic("ods2_strategy: spec");
	if (bp->b_blkno == bp->b_lblkno) {
		if ((error =
		    VOP_BMAP(vp, bp->b_lblkno, NULL, &bp->b_blkno, NULL, NULL))) {
			bp->b_error = error;
			bp->b_flags |= B_ERROR;
			biodone(bp);

if (0)
printf("ods2_strategy[ending, error=%d]\n", error);

			return (error);
		}
		if ((long)bp->b_blkno == -1)
			clrbuf(bp);
	}
	if ((long)bp->b_blkno == -1) {
		biodone(bp);
		return (0);
	}
	vp = ip->i_devvp;
	bp->b_dev = vp->v_rdev;

if (0)
printf("ods2_strategy[calling VOP_STRATEGY]\n");

	VOP_STRATEGY(vp, bp);

if (0)
printf("ods2_strategy[ending]\n");

	return (0);
}

/*
 * Print out the contents of an inode.
 */
static int
ods2_print(ap)
	struct vop_print_args /* {
		struct vnode *a_vp;
	} */ *ap;
{

if (0)
	printf("tag VT_ISOFS, isofs vnode\n");
	return (0);
}

/*
 * Return POSIX pathconf information applicable to ods2 filesystems.
 */
static int
ods2_pathconf(ap)
	struct vop_pathconf_args /* {
		struct vnode *a_vp;
		int a_name;
		register_t *a_retval;
	} */ *ap;
{
if (0)
printf("ods2_pathconf[starting]\n");

	switch (ap->a_name) {
	case _PC_LINK_MAX:
		*ap->a_retval = 1;
		return (0);
	case _PC_NAME_MAX:
                *ap->a_retval = 86; /* name.ext;version */
		return (0);
	case _PC_PATH_MAX:
		*ap->a_retval = PATH_MAX;
		return (0);
	case _PC_PIPE_BUF:
		*ap->a_retval = PIPE_BUF;
		return (0);
	case _PC_CHOWN_RESTRICTED:
		*ap->a_retval = 1;
		return (0);
	case _PC_NO_TRUNC:
		*ap->a_retval = 1;
		return (0);
	default:
		return (EINVAL);
	}
	/* NOTREACHED */
}

/*
 * get page routine
 *
 * XXX By default, wimp out... note that a_offset is ignored (and always
 * XXX has been).
 */
int
ods2_getpages(ap)
	struct vop_getpages_args *ap;
{
	return vnode_pager_generic_getpages(ap->a_vp, ap->a_m, ap->a_count,
		ap->a_reqpage);
}

/*
 * put page routine
 *
 * XXX By default, wimp out... note that a_offset is ignored (and always
 * XXX has been).
 */
int
ods2_putpages(ap)
	struct vop_putpages_args *ap;
{
	return vnode_pager_generic_putpages(ap->a_vp, ap->a_m, ap->a_count,
		ap->a_sync, ap->a_rtvals);
}

/*
 * Last reference to an inode, write the inode out and if necessary,
 * truncate and deallocate the file.
 */
static int
ods2_inactive(ap)
        struct vop_inactive_args /* {
                struct vnode *a_vp;
                struct proc *a_p;
        } */ *ap; 
{
        struct vnode *vp = ap->a_vp;
        struct proc *p = ap->a_p;
        register struct ods2_node *ip = VTOODS2(vp);
        int error = 0;

if (0)
printf("ods2_inactive[starting, ino=%d]\n", ip->i_num);
 
        if (prtactive && vp->v_usecount != 0)
                vprint("ods2_inactive: pushing active", vp);
 
/*        ip->i_flag = 0;  */
        VOP_UNLOCK(vp, 0, p);   
        /*
         * If we are done with the inode, reclaim it
         * so that it can be reused immediately.
         */
/*        if (ip->inode.iso_mode == 0) */ /* XXX: ? */
                vrecycle(vp, (struct simplelock *)0, p);
        return error;
}

/*
 * Reclaim an inode so that it can be used for other purposes.
 */
static int
ods2_reclaim(ap)
        struct vop_reclaim_args /* { 
                struct vnode *a_vp;
                struct proc *a_p;
        } */ *ap; 
{
        register struct vnode *vp = ap->a_vp;
        register struct ods2_node *ip = VTOODS2(vp);

if (0)
printf("ods2_reclaim[starting, ino=%d]\n", ip->i_num);
 
        if (prtactive && vp->v_usecount != 0)
                vprint("ods2_reclaim: pushing active", vp);
        /*
         * Remove the inode from its hash chain.
         */
        ods2_ihashrem(ip);
        /*
         * Purge old data structures associated with the inode.
         */
        cache_purge(vp);
        if (ip->i_devvp) {
                vrele(ip->i_devvp);
                ip->i_devvp = 0;
        }
	FREE(ip->file_header, M_ODS2FSFH);
        FREE(vp->v_data, M_ODS2FSNODE);
        vp->v_data = NULL;
        return (0);
}

static int
ods2_ioctl(ap)
	struct vop_ioctl_args /* {
		struct vnode *a_vp;
        	u_long a_command;
       		caddr_t a_data;
        	int a_fflag;
        	struct ucred *a_cred;
		struct proc *a_p;
	} */ *ap;
{
    u_long command = ap->a_command;
    caddr_t data = ap->a_data;
    struct vnode *vp = ap->a_vp;
    struct ods2_node *ip = VTOODS2(vp);

    switch ( command )
    {
        case ODS2IOCGRMS_ATTRIB :
           bcopy( &ip->file_header->fh2_w_recattr, data, RMS_ATTRIB_S_LENGTH );
           return 0;

        default :
            return ENOIOCTL;
    }

    return 0;
}

/*
 * Global vfs data structures for ods2
 */
vop_t **ods2_vnodeop_p;
static struct vnodeopv_entry_desc ods2_vnodeop_entries[] = {
	{ &vop_default_desc,		(vop_t *) vop_defaultop },
	{ &vop_access_desc,		(vop_t *) ods2_access },
	{ &vop_bmap_desc,		(vop_t *) ods2_bmap },
	{ &vop_cachedlookup_desc,	(vop_t *) vfs_cache_lookup }, /* xxx */
	{ &vop_getattr_desc,		(vop_t *) ods2_getattr },
	{ &vop_inactive_desc,		(vop_t *) ods2_inactive },
	{ &vop_islocked_desc,		(vop_t *) vop_stdislocked },
	{ &vop_lock_desc,		(vop_t *) vop_stdlock },
	{ &vop_lookup_desc,		(vop_t *) ods2_lookup },
	{ &vop_pathconf_desc,		(vop_t *) ods2_pathconf },
	{ &vop_print_desc,		(vop_t *) ods2_print },
	{ &vop_read_desc,		(vop_t *) ods2_read }, /* xxx */
	{ &vop_readdir_desc,		(vop_t *) ods2_readdir }, /* xxx */
	{ &vop_readlink_desc,		(vop_t *) vop_defaultop }, /* xxx */
	{ &vop_reclaim_desc,		(vop_t *) ods2_reclaim },
	{ &vop_setattr_desc,		(vop_t *) ods2_setattr },
	{ &vop_strategy_desc,		(vop_t *) ods2_strategy },
	{ &vop_unlock_desc,		(vop_t *) vop_stdunlock },
	{ &vop_getpages_desc,		(vop_t *) ods2_getpages },
	{ &vop_putpages_desc,		(vop_t *) ods2_putpages },
	{ &vop_ioctl_desc,		(vop_t *) ods2_ioctl },
	{ NULL, NULL }
};
static struct vnodeopv_desc ods2_vnodeop_opv_desc =
	{ &ods2_vnodeop_p, ods2_vnodeop_entries };
VNODEOP_SET(ods2_vnodeop_opv_desc);

/*
 * Special device vnode ops
 */
vop_t **ods2_specop_p;
static struct vnodeopv_entry_desc ods2_specop_entries[] = {
	{ &vop_default_desc,		(vop_t *) spec_vnoperate },
	{ &vop_access_desc,		(vop_t *) ods2_access },
	{ &vop_getattr_desc,		(vop_t *) ods2_getattr },
	{ &vop_inactive_desc,		(vop_t *) ods2_inactive },
	{ &vop_islocked_desc,		(vop_t *) vop_stdislocked },
	{ &vop_lock_desc,		(vop_t *) vop_stdlock },
	{ &vop_print_desc,		(vop_t *) ods2_print },
	{ &vop_reclaim_desc,		(vop_t *) ods2_reclaim },
	{ &vop_setattr_desc,		(vop_t *) ods2_setattr },
	{ &vop_unlock_desc,		(vop_t *) vop_stdunlock },
	{ NULL, NULL }
};
static struct vnodeopv_desc ods2_specop_opv_desc =
	{ &ods2_specop_p, ods2_specop_entries };
VNODEOP_SET(ods2_specop_opv_desc);

vop_t **ods2_fifoop_p;
static struct vnodeopv_entry_desc ods2_fifoop_entries[] = {
	{ &vop_default_desc,		(vop_t *) fifo_vnoperate },
	{ &vop_access_desc,		(vop_t *) ods2_access },
	{ &vop_getattr_desc,		(vop_t *) ods2_getattr },
	{ &vop_inactive_desc,		(vop_t *) ods2_inactive },
	{ &vop_islocked_desc,		(vop_t *) vop_stdislocked },
	{ &vop_lock_desc,		(vop_t *) vop_stdlock },
	{ &vop_print_desc,		(vop_t *) ods2_print },
	{ &vop_reclaim_desc,		(vop_t *) ods2_reclaim },
	{ &vop_setattr_desc,		(vop_t *) ods2_setattr },
	{ &vop_unlock_desc,		(vop_t *) vop_stdunlock },
	{ NULL, NULL }
};
static struct vnodeopv_desc ods2_fifoop_opv_desc =
	{ &ods2_fifoop_p, ods2_fifoop_entries };

VNODEOP_SET(ods2_fifoop_opv_desc);
